#include <AIV/Helpers.mqh>
#include <Math/Stat/Math.mqh>

class CSets
{
  public: 

    CHelpers * helpers;

    CSets(void)
    {
      helpers = new CHelpers();
    }

    ~CSets(void)
    {
      delete(helpers);
    }

    // Get last time bar in a series
    // @param series
    datetime getTime(SSets &series[])
    {
      datetime startTime = 0;
      int count = ArraySize(series);
  
      for (int k = 0; k < count; k++)
      {
        startTime = MathMax(startTime, (int) SymbolInfoInteger(series[k].name, SYMBOL_TIME));
      }
  
      return startTime;
    }
    
    // Get number of minimum available bars in history
    // @param series
    // @param period
    int getBars(SSets& series[], const ENUM_TIMEFRAMES period = PERIOD_CURRENT)
    {
      int size = ArraySize(series);
      int bars = TerminalInfoInteger(TERMINAL_MAXBARS);

      for (int k = 0; k < size; k++)
      {
        bars = MathMin(bars, Bars(series[k].name, period));

        if (bars < 1 || SeriesInfoInteger(series[k].name, period, SERIES_SYNCHRONIZED) == 0)
        {
          return 0;
        }
      }

      return bars;
    }

    // Split string of symbols by separator
    // @param series
    // @param names
    // @param validate
    int getSymbols(
      SSets& series[], 
      const string names, 
      const bool validate = false)
    {
      string syms[];
      int index = 0;
      int count = StringSplit(names, ',', syms);

      for (int k = 0; k < count; k++)
      {
        bool isValid = 
          SymbolSelect(syms[k], true) && 
          SymbolInfoInteger(syms[k], SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL;

        if (isValid || validate == false)
        {
          ArrayResize(series, index + 1); 
          series[index].name = syms[k];
          index++;
        }
      }
      
      return index;
    }

    // List combinations of provided currencies
    // @param series
    // @param names
    // @param end
    int getPairs(
      SSets& series[], 
      const string names, 
      const string end = "")
    {
      string syms[];

      int index = 0;
      int count = StringSplit(names, ',', syms);

      for (int k = 0; k < count; k++)
      {
        for (int n = k + 1; n < count; n++)
        {
          string name = syms[k] + syms[n] + end;

          if (SymbolSelect(name, true) == false)
          {
            name = syms[n] + syms[k] + end;
          }

          if (SymbolInfoInteger(name, SYMBOL_TRADE_MODE) == SYMBOL_TRADE_MODE_FULL)
          {
            ArrayResize(series, index + 1);
            series[index].name = name;
            index++;
          }
        }
      }

      return index;
    }

    // Get quotes without synchronization
    // @param series
    // @param position
    // @param count
    int getQuotes(
      SSets& series[], 
      const int position, 
      const int count,
      const ENUM_TIMEFRAMES period = PERIOD_CURRENT)
    {
      int items = 1000000;
      int size = ArraySize(series);

      for (int k = 0; k < size; k++)
      {
        items = MathMin(items, helpers.getQuotes(series[k], position, count, period));

        if (items < 1)
        {
          return 0;
        }
      }

      return items;
    }

    // Get list of quotes synchronized by date
    // @param series
    // @param position
    // @param count
    int getSyncQuotes(
      SSets &series[], 
      const int position = 0, 
      const int count = 0,
      ENUM_TIMEFRAMES period = PERIOD_CURRENT) 
    {
      int indexes[];
      int items = 10000;
      int size = ArraySize(series);

      SSets sources[];

      ArrayResize(sources, size);
      ArrayResize(indexes, size);
      ZeroMemory(indexes);

      for (int k = 0; k < size; k++)
      {
        indexes[k] = 0;
        items = MathMin(items, CopyRates(series[k].name, period, position, count, sources[k].rates));

        if (items < 1)
        {
          Print(
            "Synchronization : " + series[k].name + ", " + 
            "Position : " + IntegerToString(position) + ", " + 
            "Depth : " + IntegerToString(count));
  
          return 0;
        }
      }

      for (int k = 0; k < size; k++)
      {
        ArraySetAsSeries(series[k].rates, true);
        ArraySetAsSeries(sources[k].rates, true);
        ArrayResize(series[k].rates, items);
        ArrayResize(sources[k].rates, items);
      }

      for (int n = 0; n < items; n++) 
      {
        datetime pointTime = 0;

        for (int k = 0; k < size; k++)
        {
          pointTime = MathMax(pointTime, sources[k].rates[n].time);
        }

        for (int k = 0; k < size; k++)
        {
          MqlRates currentRate = sources[k].rates[indexes[k]];

          if (currentRate.time >= pointTime) 
          {
            indexes[k]++;
          }

          series[k].rates[n] = currentRate;
          series[k].rates[n].time = pointTime;
        }
      }

      return items;
    }

    int getSyncTicks(
      SSets &series[], 
      const ulong start = 0, 
      const ulong end = 0) 
    {
      int indexes[];
      int items = 10000;
      int size = ArraySize(series);

      SSets sources[];

      ArrayResize(sources, size);
      ArrayResize(indexes, size);
      ZeroMemory(indexes);

      for (int k = 0; k < size; k++)
      {
        indexes[k] = 0;
        items = MathMin(items, CopyTicksRange(series[k].name, sources[k].ticks, COPY_TICKS_ALL, start, end));
        //items = MathMin(items, CopyTicks(series[k].name, sources[k].ticks, COPY_TICKS_ALL, 0, 1000));

        if (items < 1)
        {
          return 0;
        }
      }

      for (int k = 0; k < size; k++)
      {
        ArraySetAsSeries(series[k].ticks, true);
        ArraySetAsSeries(sources[k].ticks, true);
        ArrayResize(series[k].ticks, items);
        ArrayResize(sources[k].ticks, items);
      }

      datetime nextTime = TimeCurrent() + 1;

      for (int n = 0; n < items - 1; n++) 
      {
        datetime pointTime = 0;

        for (int k = 0; k < size; k++)
        {
          pointTime = MathMax(pointTime, sources[k].ticks[n].time);
        }

        pointTime = MathMin(pointTime, nextTime);        
        nextTime = pointTime - 1;

        for (int k = 0; k < size; k++)
        {
          MqlTick currentRate = sources[k].ticks[indexes[k]];

          if (currentRate.time >= pointTime) 
          {
            indexes[k]++;
          }

          series[k].ticks[n] = currentRate;
          series[k].ticks[n].time = pointTime;
        }
      }

      return items;
    }

    // Get custom symbol neutral
    // @param names
    int getNeutralCombinations(const string names)
    {
      struct SCombo
      {
        string up[];
        string down[];
      } groups[];

      string combos[];
      string symbols[];

      int count = StringSplit(names, ',', symbols);
      int size = (int) MathPow(2, count);
      int comboIndex = 0;

      ArrayResize(groups, size + 1);

      for (int sizeIndex = 1; sizeIndex < size; sizeIndex++)
      {
        int bits = 0;
        int bitsIndex = 0;
        int valueIndex = 0;
        int value = sizeIndex;

        while (value > 0)
        {
          bits += (value & 1);
          value >>= 1;
        }

        ArrayResize(groups[sizeIndex].up, bits);
        ArrayResize(groups[sizeIndex].down, bits);

        value = sizeIndex;

        while (value > 0)
        {
          if ((value & 1) == 1)
          {
            groups[sizeIndex].up[bitsIndex] = symbols[valueIndex];
            groups[sizeIndex].down[bitsIndex] = symbols[valueIndex];
            bitsIndex++;
          }

          value >>= 1;
          valueIndex++;
        }

        for (int k = 0; k < bits; k++)
        {
          for (int n = 0; n < bits; n++)
          {
            if (k != n)
            {
              string symbol = groups[sizeIndex].down[k];
              groups[sizeIndex].down[k] = groups[sizeIndex].down[n];
              groups[sizeIndex].down[n] = symbol;
            }
          }

          string combo = "";

          for (int n = 0; n < bits; n++)
          {
            if (groups[sizeIndex].up[n] == groups[sizeIndex].down[n])
            {
              combo = "";
              break;
            }

            string name1 = 
              groups[sizeIndex].up[n] + ":" + 
              groups[sizeIndex].down[n];

            string name2 = 
              groups[sizeIndex].down[n] + ":" + 
              groups[sizeIndex].up[n];

            if (StringFind(combo, name1) >= 0 || StringFind(combo, name2) >= 0)
            {
              combo = "";
              break;
            }

            combo += name1 + (n < bits - 1 ? "#" : "");
          }

          if (combo != "" && (comboIndex == 0 || combos[comboIndex - 1] != combo))
          {
            ArrayResize(combos, comboIndex + 1);
            combos[comboIndex] = combo;
            comboIndex++;
            Print(combo);
          }
        }
      }

      return size;
    }

    // Get custom symbol with minimal variance
    // @param names
    // @param bars
    uint getCombinations(
      SSets &symbols[], 
      const uint bars,
      const ENUM_TIMEFRAMES period = PERIOD_CURRENT)
    {
      struct SCase
      {
        uint signs;
        double variance;
      };

      struct SSym
      {
        uint index;
        string name;
      };

      struct SCombo
      {
        SSym syms[];
        SCase cases[];
      };

      SCombo combos[];

      uint count = ArraySize(symbols);
      uint symCount = (uint) MathPow(2, count);
      uint symCases = 0;

      string combosContent = "";

      ArrayResize(combos, symCount - 1);
      ZeroMemory(combos);

      if (getSyncQuotes(symbols, 0, bars, period) < 1)
      {
        return 0;
      }

      double prices[];

      ArrayResize(prices, bars);

      for (uint comboIndex = 1; comboIndex < symCount; comboIndex++)
      {
        uint bits = 0;
        uint bitsIndex = 0;
        uint valueIndex = 0;
        uint value = comboIndex;

        while (value > 0)
        {
          bits += (value & 1);
          value >>= 1;
        }

        ArrayResize(combos[comboIndex - 1].syms, bits);

        value = comboIndex;

        while (value > 0)
        {
          if ((value & 1) == 1)
          {
            combos[comboIndex - 1].syms[bitsIndex].index = valueIndex;
            combos[comboIndex - 1].syms[bitsIndex].name = symbols[valueIndex].name;
            bitsIndex++;
          }

          value >>= 1;
          valueIndex++;
        }

        // Generate possible cases with signs

        uint caseCount = (uint) MathPow(2, bitsIndex);

        ArrayResize(combos[comboIndex - 1].cases, caseCount - 1);

        for (uint caseIndex = 1; caseIndex < caseCount; caseIndex++)
        {
          uint signIndex = 0;
          uint signValue = caseIndex;

          ZeroMemory(prices);

          while (signValue > 0)
          {
            char sign = (signValue & 1) == 1 ? 1 : -1;
            uint index = combos[comboIndex - 1].syms[signIndex].index;

            for (uint k = 0; k < bars; k++) 
            {
              prices[k] += symbols[index].rates[k].low * sign;
            }

            signValue >>= 1;
            signIndex++;
          }

          combos[comboIndex - 1].cases[caseIndex - 1].signs = caseIndex;
          combos[comboIndex - 1].cases[caseIndex - 1].variance = MathStandardDeviation(prices);

          for (uint k = caseIndex - 1; k > 0; k--) 
          {
            if (combos[comboIndex - 1].cases[k].variance < combos[comboIndex - 1].cases[k - 1].variance)
            {
              uint saveCombo = combos[comboIndex - 1].cases[k - 1].signs;
              double saveVar = combos[comboIndex - 1].cases[k - 1].variance;
              combos[comboIndex - 1].cases[k - 1].variance = combos[comboIndex - 1].cases[k].variance;
              combos[comboIndex - 1].cases[k].variance = saveVar;
              combos[comboIndex - 1].cases[k - 1].signs = combos[comboIndex - 1].cases[k].signs;
              combos[comboIndex - 1].cases[k].signs = saveCombo;
            }
          }
        }

        symCases += caseCount;
      }

      uint combosCount = ArraySize(combos);

      for (uint comboIndex = 0; comboIndex < combosCount; comboIndex++)
      {
        string comboName = "#";

        uint symsCount = ArraySize(combos[comboIndex].syms);

        for (uint symIndex = 0; symIndex < symsCount; symIndex++)
        {
          comboName += combos[comboIndex].syms[symIndex].name + "#";
        }

        string comboContent = 
          comboName + " : " + 
          DoubleToString(combos[comboIndex].cases[0].signs, 5) + " : " + 
          DoubleToString(combos[comboIndex].cases[0].variance, 5);

        combosContent += comboContent + "\n";
      }

      helpers.debug(combosContent);
      return symCases;
    }
};